# -*- coding: utf-8 -*-
# ===============================================
# (C) 2018 AO Kaspersky Lab. All Rights Reserved.
# ===============================================
import sys
import os, inspect
import base64
import gzip
import cStringIO as StringIO

if sys.platform == 'linux2':
    packages_path = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
    packages_path = os.path.join(packages_path, "dt_packages")
    sys.path.append(packages_path)

import json
from pandas import DataFrame, Series
import pandas as pd
import numpy as np
from sklearn.preprocessing import LabelEncoder
from sklearn.preprocessing import OneHotEncoder
from sklearn.feature_extraction.text import CountVectorizer
import xgboost as xgb
import dt_vhs_predictor
import pickle
import scipy
from datetime import datetime, timedelta
import logging
import re
import dt_Options as Options

import dt_Data as Data
from dt_classifier_common import is_ipv4, is_ipv6, is_local
from dt_classifier_common import is_md5_up_to_date, is_vhs_up_to_date, is_path_up_to_date, is_domain_data_up_to_date
import dt_PathConfig as PathConfig
import traceback




COMMENT_RARE_DOMAIN = "Domain is unpopular in the corporate network."
COMMENT_RECENT_DOMAIN = "Domain appeared in the corporate network less than {interval} ago."
COMMENT_RARE_COUNTRY = "Country code top-level domain is rarely visited in the corporate network."
COMMENT_RARE_NETWORK = "Domain belongs to a network, which is rarely visited in the company."





def evaluate_vhs(data):
    if Options.VHS_DETECT_ENABLED:
        logging.info('Vhs evaluation')
    else:
        logging.info('Vhs evaluation disabled')

    data.file['HDR']['VhsEvaluate'] = len(data.file['HDR'])

    vhs_num = data.file['HDR']['VhsEvaluate']

    predict_buf = []
    file_lines = []
    for linePair in data.file.items():
        if linePair[0] in ('HDR', 'KEYS'):
            continue;
        linePair[1].append(None)
        vhs_line = data.GetValueInLine(data.file, linePair[1], 'VhsDataBlob')
        if vhs_line is None or len(vhs_line)<815:
            continue;
        file_lines.append(linePair[1]);
        predict_buf.append(vhs_line)

    if Options.VHS_DETECT_ENABLED and len(predict_buf) >0:
        vhs_predictor = dt_vhs_predictor.VhsPredictor()
        VHS_MODEL_PATH = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
        VHS_MODEL_PATH = os.path.join(VHS_MODEL_PATH, "data", "dt_vhs_model.dat")
        vhs_predictor.Load(VHS_MODEL_PATH)
        vhs_predictor._predictor._Booster.set_param({'nthread' : 1})
        vhs_results = vhs_predictor.PredictBlobs(predict_buf, searchNeighbors=False)

        vhs_stat = {}
        for i in range(0, len(file_lines)):
            vhs_result = vhs_results['verdict'][i]
            file_lines[i][vhs_num] = vhs_result

            if vhs_result in vhs_stat:
                vhs_stat[vhs_result] += 1
            else:
                vhs_stat[vhs_result] = 1

        logging.info('Vhs results:' + str(vhs_stat))

    return len(file_lines)



def CheckPath(path):
    if path is None:
        return 0, ""

    total_suspicion = 0
    reasons = []
    for key, value in PathConfig.SuspicionPaths.iteritems():
        regexp = value['regexps']
        suspicion = value['suspicion']

        if regexp.match(path) is not None:
            total_suspicion += suspicion
            reasons.append("{0} path rule alert".format(key))

    return total_suspicion, "\n".join(reasons)


def evaluate_path_regex(data):
    logging.info('Evaluating path regex')

    data.path['HDR']['pathevaluateregex'] = len(data.path['HDR'])

    path_num = data.path['HDR']['pathevaluateregex']

    num = 0
    for line_pair in data.path.iteritems():
        if line_pair[0] in ('HDR', 'KEYS'):
            continue;
        file_path_norm = line_pair[0].decode('utf-8')
        line_pair[1].append(CheckPath(file_path_norm))
        num+= 1

    return num



from dt_haffmancodelength import HaffmanCodeLength
from dt_haffmanstringdefiner import HaffmanStringDefiner
from dt_pathstrdefiner import PopPathChecker

skip_ext_list = [u'.bat', u'.cmd', u'.vbs', u'.js', u'.ps']

def evaluate_path(data):
    logging.info('Evaluating path: loading models')
    MODEL_PATH = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
    train_columns, encoding_columns, le_dict, ohe, count_vect_words, count_vect_path_words, count_vect_sub_path, model = pickle.load(open(os.path.join(MODEL_PATH, "data",'dt_path_model.dat'), "rb"))
    pc = PopPathChecker(os.path.join(MODEL_PATH, "data",'dt_path_model_pop_dict.dat'))
    hw = HaffmanStringDefiner(HaffmanStringDefiner.load_dict_from_file(os.path.join(MODEL_PATH, "data","dt_path_model_words_count.dat")))

    logging.info('Evaluating path: prepare data 1-stage')
    data.path['HDR']['pathevaluate'] = len(data.path['HDR'])
    data.path['HDR']['full_path_nrnd'] = len(data.path['HDR'])

    path_num = data.path['HDR']['pathevaluate']
    full_path_nrnd_num = data.path['HDR']['full_path_nrnd']

    file_path_dict = {}
    file_path_index = []
    file_path_features_dict = {}
    file_path_features_dict['fname_users_cnt'] = []
    file_path_features_dict['full_path_users_cnt'] = []
    file_path_features_dict['short_path_users_cnt'] = []
    for line_pair in data.path.iteritems():
        if line_pair[0] in ('HDR', 'KEYS'):
            continue;
        file_path_norm = line_pair[0].decode('utf-8')
        line_pair[1].append(None)
        line_pair[1].append(None)
        file_path_dict[file_path_norm] = line_pair[1]

        if (data.GetValueInLine(data.path, line_pair[1], 'UpdateTimeFilePathNorm') is None
            or data.GetValueInLine(data.path, line_pair[1], 'UpdateTimeFileName') is None):
            continue
        file_path_norm_l = file_path_norm.lower()
        skip_by_ext = False
        for ext in skip_ext_list:
            if file_path_norm_l.endswith(ext):
                skip_by_ext = True
                break
        if skip_by_ext:
            continue

        file_path_features_dict['fname_users_cnt'].append(data.GetValueInLine(data.path, line_pair[1], 'FileNameUsersCount'))
        file_path_features_dict['full_path_users_cnt'].append(data.GetValueInLine(data.path, line_pair[1], 'UsersCount'))
        file_path_features_dict['short_path_users_cnt'].append(data.GetValueInLine(data.path, line_pair[1], 'ShortPathNormUsersCount'))
        file_path_index.append(file_path_norm)



    if len(file_path_index)==0:
        logging.info('Evaluating path: no path for check')
        return 0

    logging.info('Evaluating path: prepare data 2-stage')
    df = DataFrame(file_path_features_dict, index = file_path_index)

    df = pd.concat([df, df.apply(pc.parse_full_path_norm, axis=1)], axis = 1)
    df = pd.concat([df,
                    df.apply(pc.get_pop_file_path, axis=1),
                    df.apply(pc.get_pop_file_name, axis=1),
                    df.apply(hw.get_haffman_weight, axis=1)
                    ] , axis = 1)

    logging.info('Evaluating path: prepare data 3-stage')
    df['PopFilePath'] = df['PopFilePath'].fillna(df['file_path_template_code'])
    df['afterPopPathSubFolders'] = df['afterPopPathSubFolders'].fillna(0)
    df['popPathWordsCnt'] = df['popPathWordsCnt'].fillna(0)
    df['popSubFoldersCnt'] = df['popSubFoldersCnt'].fillna(0)
    df['isPopFilePathOnly'] = df['isPopFilePathOnly'].fillna(True)
    df['popWords'] = df['popWords'].fillna('')
    df['popPathWords'] = df['popPathWords'].fillna('')
    df['popSubFolders'] = df['popSubFolders'].fillna('')
    df['isPopFilePathOnly_num'] = 0
    df.loc[df['isPopFilePathOnly'] == True, 'isPopFilePathOnly_num'] = 1

    df['fname_users_cnt'] = df['fname_users_cnt'].fillna(0)
    df['full_path_users_cnt'] = df['full_path_users_cnt'].fillna(0)
    df['short_path_users_cnt'] = df['short_path_users_cnt'].fillna(0)

    num_columns = [
        'fname_users_cnt'
        ,'full_path_users_cnt'
        ,'short_path_users_cnt']
    for column in num_columns:
        df.loc[df[column]<10, column]=0

    df_sp1 = scipy.sparse.csr.csr_matrix(df[train_columns].values)
    encoding_columns_num = []
    for n in encoding_columns:
        df[n] = df[n].fillna('')
        replace = ''
        if n=='PopFilePath':
            replace = '0'
        df[n] = df[n].map(lambda s: replace if s not in le_dict[n].classes_  else s)
        df[n+'_num'] = le_dict[n].transform(df[n])
        encoding_columns_num.append(n+'_num')
    df_sp2 = ohe.transform(df[encoding_columns_num])

    df_sp3 = count_vect_words.transform(df['popWords'])
    df_sp4 = count_vect_path_words.transform(df['popPathWords'])
    df_sp5 = count_vect_sub_path.transform(df['popSubFolders'])

    data_xx = scipy.sparse.hstack(
        (df_sp1, df_sp2, df_sp3, df_sp4, df_sp5)
        , format='csr')

    logging.info('Evaluating path: predict by model')
    all_score = model.predict_proba(data_xx)[:,1]

    df['score']= all_score

    bad_cnt = df[df['score']>0.994792163372].shape[0]
    good_cnt = df[df['score']<0.0195892006159].shape[0]
    total_cnt = df.shape[0]

    logging.info('Evaluating path: save to general data. Has bad:'+ str(bad_cnt)+' good:'+str(good_cnt)+' all:'+str(total_cnt))

    for key, value in df.iterrows():
        if key is None:
            logging.warning('None path key')
            continue
        if not file_path_dict.has_key(key):
            logging.warning('No path key: '+key)
            continue
        line = file_path_dict[key]
        line[path_num] = value['score']
        line[full_path_nrnd_num] = value['full_path_nrnd'].encode('utf-8')


    return df.shape[0]




def get_host_criteria(totalCompCount):
    return {
        "NetLiveLongDays": {
            "Low": {
                "threshold": 14,
                "suspicion": 20,
                "comment": COMMENT_RECENT_DOMAIN.format(interval="a week")
            },
            "High": {
                "threshold": 30,
                "suspicion": -100}
        },
        "CountryNetCompCount":{
            "Low": {
                "threshold": min(int(totalCompCount * 0.02), 20),
                "suspicion": 0,
                "comment": COMMENT_RARE_COUNTRY
            }
        },


        "DomainNetCompCount": {
            "Low": {
                "threshold": min(int(totalCompCount * 0.01), 5),
                "suspicion": 20,
                "comment": COMMENT_RARE_DOMAIN
            },
            "High":{
                "threshold": 10,
                "suspicion": -20}
        },
        "NetNameNetCompCount":{
            "Low":{
                "threshold": min(int(totalCompCount * 0.01), 5),
                "suspicion": 20,
                "comment": COMMENT_RARE_NETWORK
            },
            "High":{
                "threshold": 20,
                "suspicion": -20}
        },
        "OrgNameNetCompCount":{
            "High": {
                "threshold": 50,
                "suspicion": -20}
        },
        "OwnerNameNetCompCount":{
            "Low": {
                "threshold": min(int(totalCompCount * 0.01), 5),
                "suspicion": 20,
                "comment": COMMENT_RARE_NETWORK
            },
            "High":{
                "threshold": 50,
                "suspicion": 0}
        },
        "BestRank": {
            "Low": {
                "threshold": 100000,
                "suspicion": -100},
            "Low2": {
                "threshold": 1000000,
                "suspicion": -30}

        },
        "LiveLongDays": {
            "Low": {
                "threshold": 50,
                "suspicion": 20,
                "comment": COMMENT_RECENT_DOMAIN.format(interval="365 days")},
            "High": {
                "threshold": 365,
                "suspicion": -40}
        },
        "IsDomain": {
            "High": {
                "threshold": 0,
                "suspicion": 0}
        }
    }


def get_domain_best_rank(domainInfo):
    ranks = [domainInfo['DomainShortKsnRank'],
             domainInfo['RemoteHostKsnRankMin']]

    domain_short_ksn_users = domainInfo['DomainShortKsnUsers'] or 0
    remote_host_ksn_users_max = domainInfo['RemoteHostAKsnUsersMax'] or 0
    if domain_short_ksn_users > 500 or remote_host_ksn_users_max > 500:
        ranks.append(100)

    ranks = filter(None, ranks)
    return min(ranks) if ranks else None


def get_domain_info(domain, data, info_to_load):
    domainInfo = Data.CaseInsensitiveDict()
    curDomain = data.domain[domain]
    for x in info_to_load:
        domainInfo[x] = data.GetValueInLine(data.domain, curDomain, x)
    return domainInfo


TRUSTED_OWNERS = ['kaspersky', 'microsoft', 'apple', 'google', 'symantec', 'skype', 'akamai technologies', 'telegram messenger']

def is_trusted_owner(owner):
    if owner is None or owner =='':
        return False
    l_owner = owner.lower()
    for trusted in TRUSTED_OWNERS:
        if trusted in l_owner:
            return True
    return False

def get_host_hazard(domain, data, net_comp_count):
    infoToLoad=['DomainShortKsnRank',
                'RemoteHostKsnRankMin',
                'DomainShortKsnUsers',
                'RemoteHostAKsnUsersMax',
                'CreationDate',
                'DomainIp',
                'CountryNetCompCount',
                'OrgNameNetCompCount',
                'NetNameNetCompCount',
                'OwnerNameNetCompCount',
                'OrgName',
                'NetName',
                'OwnerName',
                'CountryName',
                'DomainNetCompCount',
                'DomainNetFirstTime',
                'OwnerName',
                'NetName',
                'OrgName',
                'Category',
                'Zone',
                'Descr']

    if is_ipv6(domain):
        return (-1,  'IpV6 Domain.', 'IpV6 Domain.')

    if is_local(domain):
        return (-1,  'LocalIP Domain.', 'LocalIP Domain.')

    domainUpToDate, upd_info = is_domain_data_up_to_date(domain, data)
    if not domainUpToDate:
        return (-1,  'No actual data for ' + upd_info, 'No actual data for domain')

    domainInfo = get_domain_info(domain, data, infoToLoad)

    if domainInfo['DomainIp'] is not None and is_local(domainInfo['DomainIp']):
        return (-1,  'LocalIP Domain.',  'LocalIP Domain.')

    if (is_trusted_owner(domainInfo['OrgName']) or
        is_trusted_owner(domainInfo['NetName']) or
        is_trusted_owner(domainInfo['OwnerName']) or
        is_trusted_owner(domainInfo['Descr'])):
        return (0,  'Trusted owner', 'Trusted owner')

    if not domainInfo['OrgName'] or domainInfo['OrgNameNetCompCount'] == 0:
        domainInfo['OrgNameNetCompCount'] = None

    if not domainInfo['NetName'] or domainInfo['NetNameNetCompCount'] == 0:
        domainInfo['NetNameNetCompCount'] = None

    if not domainInfo['CountryName'] or domainInfo['CountryNetCompCount'] == 0:
        domainInfo['CountryNetCompCount'] = None

    if not domainInfo['OwnerName'] or domainInfo['OwnerName'] == '' or domainInfo['OwnerNameNetCompCount'] == 0:
        domainInfo['OwnerNameNetCompCount'] = None

    if not is_ipv4(domain):
        domainInfo['IsDomain'] = 1

    domainInfo['NetLiveLongDays'] = None
    if domainInfo['DomainNetFirstTime'] is not None:
        domainInfo['NetLiveLongDays'] = (data.to_time - domainInfo['DomainNetFirstTime']).days

    domainInfo['LiveLongDays'] = None
    if domainInfo['CreationDate'] is not None:
        domainInfo['LiveLongDays'] = (data.to_time - domainInfo['CreationDate']).days

    domainInfo['BestRank'] = get_domain_best_rank(domainInfo)

    suspicion = 0
    reasons = []
    comments = []
    for criterion, rules in get_host_criteria(net_comp_count).iteritems():
        domain_value = domainInfo.get(criterion)
        if domain_value is None:
            continue

        rule = rules.get('Low')
        if rule is not None and domain_value < rule['threshold']:
            suspicion += rule["suspicion"]
            reasons.append("{0} ({1}) is less than {2}".format(criterion, domain_value, rule['threshold']))

            comment = rule.get('comment')
            if comment is not None:
                comments.append(comment)

        rule = rules.get('Low2')
        if rule is not None and domain_value < rule['threshold']:
            suspicion += rule["suspicion"]
            reasons.append("{0} ({1}) is less than {2}".format(criterion, domain_value, rule['threshold']))

            comment = rule.get('comment')
            if comment is not None:
                comments.append(comment)

        rule = rules.get('High')
        if rule is not None and domain_value >= rule["threshold"]:
            suspicion += rule["suspicion"]
            reasons.append("{0} ({1}) is greater than {2}".format(criterion, domain_value, rule["threshold"]))

            comment = rule.get('comment')
            if comment is not None:
                comments.append(comment)

    if suspicion > 70:
        result = 2
    elif suspicion > 50:
        result = 1
    else:
        result = 0

    reasons.sort()
    comments.sort()

    return result, "\n".join(reasons), "\n".join(set(comments))



def evaluate_domain(data):
    net_comp_count = len(data.comp) - 2
    data.domain['HDR']['domain_hazard'] = len(data.domain['HDR'])
    line_num = data.domain['HDR']['domain_hazard']

    dict_results = {}

    for line_pair in data.domain.iteritems():
        if line_pair[0] in ('HDR', 'KEYS'):
            continue;
        line_pair[1].append(None)
        domain = line_pair[0]
        result = get_host_hazard(domain, data, net_comp_count)
        line_pair[1][line_num] = result
        txt = str(result[0]) + ' - ' + result[2]
        if not dict_results.has_key(txt):
            dict_results[txt] = 0
        else:
            dict_results[txt] +=1

    res_text = ''
    lst_results = list(dict_results.items())
    for x in lst_results:
        res_text+= '\n   ' + str(x[0])+': '+str(x[1])

    logging.info('Evaluating domain results: '+res_text)
    return len(data.domain) - 2



def get_app_is_wl(file, path, data):

    if not file:
        return None, "No file"

    if file == '00000000000000000000000000000000':
        return None, "Zerro md5"

    hips_file_group = data.GetValue(data.file, file, 'HIPSFileGroup')
    if hips_file_group == 0:
        return True, "Trusted HIPS code"

    vhs_result = data.GetValue(data.file, file, 'VhsEvaluate')
    if vhs_result == 'Clean':
        return True, "Vhs predictor clean"

    is_sig_valid = data.GetValue(data.file, file, 'IsSignatureValid')
    if is_sig_valid == True:
        return True, "File has valid signature"


    file_vhs = data.GetValue(data.file, file, 'VHS')
    if not file_vhs is None:
        hack_tool_info = data.GetValue(data.vhs, file_vhs, 'HackToolVerdictes')
        if not hack_tool_info is None:
            hack_tool_info = hack_tool_info.lower()
            if 'whitelistedfile' in hack_tool_info:
                return True, "Vhs whitelisted file"

    return False, ''


def get_app_is_pop(file, path, data):

    if not file:
        return None, "No file"

    if file == '00000000000000000000000000000000':
        return None, "Zerro md5"

    file_vhs = data.GetValue(data.file, file, 'VHS')
    md5_up_to_date = is_md5_up_to_date(file, data)
    path_up_to_date = is_path_up_to_date(path, data)

    if not md5_up_to_date:
        return None, "No actual data for md5"

    if not path_up_to_date:
        return None, "No actual data for path"

    if not file_vhs is None:
        vhs_up_to_date = is_vhs_up_to_date(file_vhs, data)
        if vhs_up_to_date == False:
            return None, "No actual data for vhs"

    user_count_ksn_by_md5 = data.GetValue(data.file, file, 'UsersCount')
    if user_count_ksn_by_md5 > 500:
        return True, "File md5 has {0} KSN users".format(user_count_ksn_by_md5)

    user_count_ksn_by_path = data.GetValue(data.path, path, 'UsersCount')
    if user_count_ksn_by_path > 500:
        return True, "File path has {0} KSN users".format(user_count_ksn_by_path)

    if not file_vhs is None:
        user_count_ksn_by_vhs = data.GetValue(data.vhs, file_vhs, 'UsersCount')
        if vhs_up_to_date and user_count_ksn_by_vhs > 500:
            return True, "VHS has {0} KSN users".format(user_count_ksn_by_vhs)

    comp_count_net_by_md5 = data.GetValue(data.file, file, 'NetCompCount')
    if comp_count_net_by_md5 > 50:
        return True, "Has {0} net comp count".format(comp_count_net_by_md5)

    comp_count_net_by_path = data.GetValue(data.path, path, 'NetCompCount')
    if path is not None and comp_count_net_by_path > 50:
        return True, "Path has on {0} comps in net".format(comp_count_net_by_path)

    return False, ''



def get_app_is_very_pop_by_ksn(file, path, data):

    if not file:
        return None, "No file"

    if file == '00000000000000000000000000000000':
        return None, "Zerro md5"

    file_vhs = data.GetValue(data.file, file, 'VHS')
    md5_up_to_date = is_md5_up_to_date(file, data)
    path_up_to_date = is_path_up_to_date(path, data)

    if not md5_up_to_date:
        return None, "No actual data for md5"

    if not path_up_to_date:
        return None, "No actual data for path"

    if not file_vhs is None:
        vhs_up_to_date = is_vhs_up_to_date(file_vhs, data)
        if vhs_up_to_date == False:
            return None, "No actual data for vhs"

    user_count_ksn_by_md5 = data.GetValue(data.file, file, 'UsersCount')
    if user_count_ksn_by_md5 > 5000:
        return True, "File md5 has {0} KSN users".format(user_count_ksn_by_md5)

    user_count_ksn_by_path = data.GetValue(data.path, path, 'UsersCount')
    if user_count_ksn_by_path > 5000:
        return True, "File path has {0} KSN users".format(user_count_ksn_by_path)

    if not file_vhs is None:
        user_count_ksn_by_vhs = data.GetValue(data.vhs, file_vhs, 'UsersCount')
        if vhs_up_to_date and user_count_ksn_by_vhs > 5000:
            return True, "VHS has {0} KSN users".format(user_count_ksn_by_vhs)

    return False, ''



def get_app_hazard_susp(file, path, data):

    app_is_wl = data.GetValue(data.app, (file, path), 'app_is_wl')
    app_is_pop = data.GetValue(data.app, (file, path), 'app_is_pop')
    if app_is_wl is None or app_is_pop is None or app_is_wl[0] is None or app_is_pop[0] is None:
        return False, 'No data', 'No data'
    if app_is_wl[0] == True:
        return False, app_is_wl[1], app_is_wl[1]
    if app_is_pop[0] == True:
        return False, app_is_pop[1], app_is_pop[1]

    clean_path_limit = 0.0195892006159
    path_evaluate = data.GetValue(data.path, path, 'pathevaluate') or 0
    if not path_evaluate is None and path_evaluate<clean_path_limit:
        return False, "Not dangerous path ML", ""

    suspicion, reason = data.GetValue(data.path, path, 'pathevaluateregex') or 0
    if suspicion < 0:
        return False, "Not dangerous path:\n{0}".format(reason), ""

    return True, "File is not very popular according to KSN", ""


app_factors = [
    get_app_is_wl,
    get_app_is_pop,
    get_app_is_very_pop_by_ksn,
    get_app_hazard_susp]

def evaluate_app(data):

    errors = set()

    for app_factor in app_factors:
        factor_name = app_factor.__name__.replace("get_", "")
        data.app['HDR'][factor_name] = len(data.app['HDR'])
        column_num = data.app['HDR'][factor_name]

        for line_pair in data.app.iteritems():
            if line_pair[0] in ('HDR', 'KEYS'):
                continue;
            file = line_pair[0][0]
            path = line_pair[0][1]

            res = None
            line_pair[1].append(res)
            try:
                res = app_factor(file, path, data)
            except:
                err_text = "Fatal error while {0} evaluating: {1} ({2}, {3})".format(app_factor.__name__, traceback.format_exc(100), file, path)
                if not err_text in errors:
                    logging.warning(err_text)
                    errors.add(err_text)

            line_pair[1][column_num] = res

    logging.info('Evaluating app finished')
    return len(data.app) - 2



#        if line_pair[1][line_num] == '': # or line_pair[1][line_num] == '[]' or line_pair[1][line_num] == '{}':











def decode_ksn_reply_inline(encoded_dict, decode_columns):
    encoded = 0
    a = encoded_dict['HDR']
    ksn_reply_col_num = encoded_dict['HDR']['ksnreply']
    for col in decode_columns:
        encoded_dict['HDR'][col] = len(encoded_dict['HDR'])

    for line_pair in encoded_dict.iteritems():
        if line_pair[0] in ('HDR', 'KEYS'):
            continue;

        for column_pair in decode_columns.items():
            line_pair[1].append(None)

        if line_pair[1][ksn_reply_col_num] is None:
            continue;
        if line_pair[1][ksn_reply_col_num] == '' or line_pair[1][ksn_reply_col_num] == '[]' or line_pair[1][ksn_reply_col_num] == '{}':
            line_pair[1][ksn_reply_col_num] = None
            continue;

        jsn = None
        try:

            jsn = json.loads(line_pair[1][ksn_reply_col_num])
        except:
            logging.warn("Can't decode "+str(line_pair[1][ksn_reply_col_num]))
            continue

        for column_pair in decode_columns.items():
            col_name = column_pair[0]
            if not col_name in jsn:
                continue
            col_type = column_pair[1]
            col_num = encoded_dict['HDR'][col_name]
            if type(col_type) is type:
                line_pair[1][col_num] = col_type(jsn[col_name])

            elif type(col_type) is str:
                if not col_type in jsn:
                    continue
                is_compessed = bool(jsn[col_type])
                array = {}
                if is_compessed:
                    array_base64 = jsn[col_name] or ''
                    array_raw = base64.b64decode(array_base64)

                    with gzip.GzipFile(fileobj=StringIO.StringIO(array_raw)) as in_gzip:
                        array_string = '{' + in_gzip.read().decode('utf-16le') + '}'
                    array = json.loads(array_string)
                else:
                    array = jsn[col_name]
                line_pair[1][col_num] = array
            else:
                raise TypeError('Unknown col_type: '+ str(col_type))


        encoded += 1
    return encoded


def decode_ksn_reply(data):
    decode_columns = {
        'domains_array':'compressed',
        'pch_full_path_array':'compressed', 'pch_short_path_array':'compressed', 'pch_file_name_array':'compressed',
        'pch_full_path_cnt':long, 'pch_short_path_cnt':long, 'pch_file_name_cnt':long,
        'pch_unpop_full_path_user_cnt_10':long, 'pch_unpop_short_path_user_cnt_10':long, 'pch_unpop_file_name_user_cnt_10':long,
        'pch_full_path_array':'pch_full_path_array_compressed',
        'pch_short_path_array':'pch_short_path_array_compressed',
        'pch_file_name_array':'pch_file_name_array_compressed',

        'pdll_full_path_array':'compressed', 'pdll_short_path_array':'compressed', 'pdll_file_name_array':'compressed',
        'pdll_full_path_cnt':long, 'pdll_short_path_cnt':long, 'pdll_file_name_cnt':long,
        'pdll_unpop_full_path_user_cnt_10':long, 'pdll_unpop_short_path_user_cnt_10':long, 'pdll_unpop_file_name_user_cnt_10':long,
        'pdll_full_path_array':'pdll_full_path_array_compressed',
        'pdll_short_path_array':'pdll_short_path_array_compressed',
        'pdll_file_name_array':'pdll_file_name_array_compressed',

        'domain_cnt': long, 'domain_cnt_pop':long, 'published_domains':int,
        'domains_array': 'compressed'
        }


    logging.info('Decode ksn reply for pathes')
    encoded_path = decode_ksn_reply_inline(data.path, decode_columns)

    decode_columns['norm_path'] = str

    logging.info('Decode ksn reply for files')
    encoded_file = decode_ksn_reply_inline(data.file, decode_columns)

    logging.info('Decoded file: '+str(encoded_file)+ ', pathes: ' + str(encoded_path))
    return encoded_file + encoded_path




import dt_factors_abstract
from dt_factors_abstract import AbstractSimpleFactor

dt_factors_abstract.FACTORS.extend([
    AbstractSimpleFactor('decode_ksn_reply', '1.0', decode_ksn_reply),
    AbstractSimpleFactor('evaluate_path', '1.0', evaluate_path),
    AbstractSimpleFactor('evaluate_path_regex', '1.0', evaluate_path_regex),
    AbstractSimpleFactor('evaluate_domain', '1.0', evaluate_domain),
    AbstractSimpleFactor('evaluate_vhs', '1.0', evaluate_vhs),
    AbstractSimpleFactor('evaluate_app', '1.0', evaluate_app)
    ])
